from .fonctions_base import create_variable, row_columns_conditions, neighbors_conditions, process_sat_output, color_conditions
from .affichage import afficher_grille
import subprocess
from random import randint, choice

### GENERER UNE GRILLE RESOLVABLE ###

def getDirectColoredNeighbors(center,n,dict_color):
    """
    Cette fonction renvoie la liste des cases voisines du centre qui ont sont colorées

    Paramètres:
    center (int) : la variable centrale dont on veut trouver les voisins
    n (int) : la dimension de la grille
    dict_color (dict(int : int)) : un dictionnaire qui renseigne la couleur de chaque case de la grille

    Renvoie:
    list(int) : la liste des variables colorées voisines de la case centrale

    """
    # Liste de tout les voisins directement adjacents (pas diagonale)
    voisins = [center -1, center+1, center-n, center+n]

    # On traite les effets de bord
    if center%n == 0:
        voisins.remove(center+1)
    if center%n == 1:
        voisins.remove(center-1)
    if center+n > n*2:
        voisins.remove(center + n)
    if center-n < 1:
        voisins.remove(center - n)

    # On filtre les variables non colorées
    for voisin in voisins:
        if dict_color[voisin] == 0: # 0 donc pas de couleurs
            voisins.remove(voisin)
    return voisins


def Génération_grille(n):
    """
    Cette fonction permet de générer une grille résolvable.

    Paramètre:
    n (int) : taille de la grille souhaitée

    Retourne:
    list(list(int)) : un grille de taille nxn dont chaque chiffre correspond à une couleur

    """
    grille_initiale = [[0 for i in range(n)] for i in range(n)]
    variables = create_variable(grille_initiale)
    
    # Donner les conditions sans celle des couleurs afin de placer les couronnes
    conditions = ""
    conditions += row_columns_conditions(variables)
    conditions += neighbors_conditions(variables)

    # Ecrire le fichier cnf
    random_var = f"{choice(choice(variables))} 0" # On fixe la première couronne au hasard pour éviter de fixer systématiquement les mêmes variables
    nb_clauses = conditions.count('0') + 1
    en_tete = f'p cnf {n**2} {nb_clauses}\n'
    with open('creation_grille.cnf','w') as file:
        file.write(en_tete + random_var + conditions)
    
    # Lancer gophersat
    path_to_gophersat = './gophersat_win64'
    command = [path_to_gophersat,'creation_grille.cnf']
    result = subprocess.run(command, capture_output=True, text=True)
    result = process_sat_output(result.stdout)
    
    # Affecter une couleur à chaque couronne déjà placée
    cpt = 1
    val_couleur = {}
    repartitions_zone = {i : [] for i in range(n+1)}

    for i in range(n):
        for j in range(n):
            if result[variables[i][j]] == 1:
                grille_initiale[i][j] += cpt
                cpt +=1    
            val_couleur[variables[i][j]] = grille_initiale[i][j]
            repartitions_zone[grille_initiale[i][j]].append(variables[i][j])

    # Affecter une couleur au reste des cases
    while len(repartitions_zone[0]) != 0:
        nb_non_coloré = len(repartitions_zone[0])
        if nb_non_coloré>1:
            candidat = repartitions_zone[0][randint(0,nb_non_coloré-1)]
        else:
            candidat = repartitions_zone[0][0]
        voisins = getDirectColoredNeighbors(candidat,n,val_couleur)
        nb_voisins = len(voisins)
        if nb_voisins > 0:
            # On choisit au hasard un voisin
            if nb_voisins > 1:
                voisin_choisi = voisins[randint(0,nb_voisins-1)]
            else:
                voisin_choisi = voisins[0]
            nouvelle_couleur = val_couleur[voisin_choisi]

            # On affecte la couleur au centre et on actualise la répartition des couleurs
            val_couleur[candidat] = nouvelle_couleur
            repartitions_zone[nouvelle_couleur].append(candidat)
            repartitions_zone[0].remove(candidat)
    return [[val_couleur[var] for var in row] for row in variables]

### RESOUDRE LA GRILLE ###

def SolveGrid(grille):
    """
    Cette fonction résoud le jeu puis l'affiche

    Paramètres:
    grille (list(list(str))) : grille avec les couleurs

    """
    # On initialise les conditions et les variables
    conditions = ""
    variables = create_variable(grille)

    # On ajoute les conditions nécessaires
    conditions += row_columns_conditions(variables)
    conditions += color_conditions(variables, grille)
    conditions += neighbors_conditions(variables)

    # Ecrire le fichier cnf
    nb_clauses = conditions.count('0')
    en_tete = f'p cnf {len(grille)**2} {nb_clauses}\n'

    with open('fichier_grille.cnf','w') as file:
        file.write(en_tete + conditions)

    # Lancer gophersat
    path_to_gophersat = './gophersat_win64'
    command = [path_to_gophersat,'fichier_grille.cnf']
    result = subprocess.run(command, capture_output=True, text=True)

    # afficher le résultat
    result = process_sat_output(result.stdout)
    afficher_grille(grille,variables,result)