
def sum_to_sat(liste_var):
    
    """
    Cette fonction convertit une liste de variables dont la somme doit être égale a 1
    en une série de conditions SAT exploitable par gophersat.

    Paramètres:
    liste_var (list) : la liste des variables dont la somme doit être égale à 1

    Retourne:
    str : une chaîne de caractères traduisant la condition en problème SAT dont le format
    est exploitable par gophersat.
    
    """
    
    n = len(liste_var)
    conditions = ""

    # Au moins une variable doit être vraie  
    union = ""
    for variable in liste_var:
        union += f"{variable} "
    union += " 0\n"
    conditions += union

    # Au plus une variable doit être vraie
    for i in range(n):
        for j in range(i+1, n):
            conditions += f"-{liste_var[i]} -{liste_var[j]} 0\n"
    
    return conditions


def create_variable(grille):
    """
    Crée une grille annexe avec les variables associées à chaque case de la grille.
    
    Paramètre:
    grille (list(list(str))) : grille d'entrée

    Retourne:
    list(list(int)) : grille des variables

    """
    n = len(grille)
    sortie = []
    cpt=1
    for i in range(n):
        row = []
        for j in range(n):
            row.append(cpt)
            cpt+=1
        sortie.append(row)
    return sortie

### CONDITION SUR LES LIGNES ET COLONNES ###

def row_columns_conditions(variables):
    """
    La fonction renvoie une chaîne de caractère qui correspont à une formulation SAT de la condition sur les lignes
    et les colonnes dans la grille. (Une couronne par ligne et par colonne).

    Paramètres:
    liste_var (list(int)) : grille des variables

    Retourne:
    str : une chaîne de caractères correspondant à la formulation SAT de la condition sur les lignes et colonnes.

    """

    conditions = ""
    n = len(variables)
    for line in variables:
        conditions += sum_to_sat(line)
    for j in range(n):
        conditions += sum_to_sat([variables[i][j] for i in range(n)])
    return conditions

### CONDITION SUR LES COULEURS ###

def get_all_colors(grille):
    """
    Extrait toutes les couleurs au sein de la grille.

    Paramètre:
    grille (list(list(str))) : grille de base à résoudre

    Retourne:
    list(str) : liste avec les couleurs dans la grille

    """
    colors = set()
    for row in grille:
        for cell in row:
            if isinstance(cell, list):  # If the cell is a list, unpack it
                colors.update(cell)
            else:
                colors.add(cell)
    return list(colors)

def color_conditions(variables, grille):
    """
    La fonction renvoie une chaîne de caractère qui correspont à une formulation SAT de la condition sur les
    couleurs dans la grille. (Une couronne par zone de couleur).

    Paramètres:
    variables (list(list(int))) : grille des variables
    grille (list(list(str))) : grille à résoudre

    Retourne:
    str : une chaîne de caractères correspondant à la formulation SAT de la condition sur les couleurs.

    """
    conditions = ''
    colors = get_all_colors(grille)
    n = len(grille)
    for color in colors:
        variables_with_coresponding_color = []
        for i in range(n):
            for j in range(n):
                if grille[i][j] == color:
                    variables_with_coresponding_color.append(variables[i][j])
        conditions += sum_to_sat(variables_with_coresponding_color)
    return conditions

def get_neighbors(matrix, row, col):
    """
    Récupère la liste des voisins d'une cellule dans une grille étant donné l'indice de la colonne et celui de la ligne

    Paramètre:
    matrix (list(list(int))) : grille des variables
    row (int) : indice de la ligne
    col (int) : indice de la colonne

    Retourne:
    list(int) : liste des variables voisines dans la grille de la case aux coordonnées spécifiées

    """
    neighbors = []
    rows, cols = len(matrix), len(matrix[0])
    for i in range(row - 1, row + 2):
        for j in range(col - 1, col + 2):
            if 0 <= i < rows and 0 <= j < cols:  # Check boundaries
                neighbors.append(matrix[i][j])
    return neighbors

### CONDITION SUR LES VOISINS ###

def neighbors_cond_to_sat(liste_var,center):

    """
    Cette liste crée une formulation SAT du problème des voisins: si la couronne se trouve au centre,
    alors aucune couronne ne peut se trouver parmis les voisins

    Paramètres:
    liste_var (list(int)) : la liste des variables voisines du centre
    center (int) : le centre

    Retourne:
    str : une chaîne de caractères traduisant la condition en problème SAT dont le format
    est exploitable par gophersat.

    """

    conditions = ''
    for cell in liste_var:
        if cell != center:
            conditions += f"-{cell} -{center} 0\n"
    return conditions


def neighbors_conditions(variables):
    """
    Prends en entrée une liste de voisins dans la grille d'une variables et la variable en question.
    La fonction renvoie une chaîne de caractère qui correspont à une formulation SAT de la condition sur les voisins
    dans la grille. (Deux couronnes ne peuvent pas être adjacentes)

    Paramètres:
    liste_var (list) : liste des voisins
    center (int) : variable centrale

    Retourne:
    str : une chaîne de caractères correspondant à la formulation SAT de la condition sur les voisins.

    """
    conditions = ''
    n = len(variables)
    for i in range(n):
        for j in range(n):
            center = variables[i][j]
            neighbors = get_neighbors(variables, i, j)
            new_cond = neighbors_cond_to_sat(neighbors,center)
            conditions += new_cond
    return conditions


### EXPLOITATION DES RESULTATS ###

def process_sat_output(input_string):
    """
    Permet de convertir la sortie du solver SAT en un dictionnaire donnant la valeur de vérité de chaque variables.

    Paramètre:
    input_string (str) : la sortie du solver gophersat

    Retourne:
    dict(int : int) : un dictionnaire faisant correspondre chaque variable à sa valeur de vérité
    
    """

    # Diviser l'entrée en lignes
    lines = input_string.splitlines()
    
    # Chercher la ligne qui commence par "v"
    for line in lines:
        if line.startswith("v "):
            # Enlever 'v' et traiter les valeurs
            elements = line.lstrip('v').strip().split()
            return {abs(int(element)) : 1 if int(element)>0 else 0 for element in elements}
        